#ifndef RSDEVELOP
#if (!defined PREFIX) || (!defined COEFF)
#error "cannot do anything without PREFIX _and_ COEFF defined"
#endif

#ifndef RS_I32 
#define RS_I32 int
#endif
#include<math.h>
#include<assert.h>
#include"generic_interface.h"

#ifndef __gmp_const 
#define __gmp_const const
#endif

typedef struct {
  int offset;
  COEFF *pol;
  int prec;
  void *cont;} __mpai_struct;
#endif
int mpai_nbmul=0;
typedef __mpai_struct mpai_t[1];
typedef __mpai_struct *mpai_ptr;
typedef __gmp_const __mpai_struct *mpai_srcptr;

//RS like definition
typedef __mpai_struct mpai_rep;
typedef mpai_rep mpai[1];
typedef mpai_rep * mpai_ref;
typedef const mpai_rep *mpai_const;
//

#include"mpai_alloc.h"
#include"mpai_cont_functions.h" 

//Global variable for default prec 

int mpai_prec=2;
void mpai_out_str(FILE *, const mpai_t);
void mpai_init(mpai_ptr op)
{
  int i;
  op->offset=0;
  op->pol=(COEFF*)MPAI_MALLOC(mpai_prec*sizeof(COEFF));
  for(i=0;i<mpai_prec;i++)
    gen_init(op->pol[i]);
  op->prec=mpai_prec;
  op->cont=NULL;
}

void mpai_clear(mpai_ptr op)
{
  int i;
  if (op){
    for(i=0;i<op->prec;i++)
      gen_clear(op->pol[i]);
    MPAI_FREE(op->pol);
    op->pol=NULL;
    MPAI_FREE(op->cont);
    op->cont=NULL;
  }
}

void mpai_set_prec(int prec)
{
  mpai_prec=prec;
}

void mpai_set_d(mpai_ptr res,const COEFF op)
{
  int i=0;
  res->prec=mpai_prec;
  res->offset=0;
  gen_set(res->pol[0],op);
  for(i=1;i<res->prec;i++)
    gen_set_d(res->pol[i],0);
  set_cont_d(res,op);
}

void mpai_set_si(mpai_ptr res,const int op)
{
  int i=0;
  res->prec=mpai_prec;
  res->offset=0;
  gen_set_d(res->pol[0],op);
  for(i=1;i<res->prec;i++)
    gen_set_d(res->pol[i],0);
  set_cont_i(res,op);
}
  
void mpai_mul_d(mpai_ptr res,const COEFF op)
{
  int i=0;
  for(;i<res->prec;i++)
    gen_mul(res->pol[i],res->pol[i],op);
  mul_cont_d(res,op);
}
  
void mpai_div_d(mpai_ptr res,const COEFF op)
{
  int i=0;
  for(;i<res->prec;i++)
    gen_div(res->pol[i],res->pol[i],op);
  mul_cont_d(res,op);
}

void mpai_set(mpai_ptr res,mpai_srcptr op)
{
  int i=0;
  res->offset=op->offset;
  for(i=op->prec;i<res->prec;i++)
    gen_clear(res->pol[i]);
  res->pol=(COEFF*)MPAI_REALLOC(res->pol,res->prec*sizeof(COEFF),op->prec*sizeof(COEFF));
  for(i=res->prec;i<op->prec;i++)
    gen_init(res->pol[i]);
  res->prec=op->prec;
  for(i=0;i<res->prec;i++)
    gen_set(res->pol[i],op->pol[i]);
  set_cont(res,op);
}

void mpai_add(mpai_ptr rop, mpai_srcptr op1, mpai_srcptr op2)
{
  int i;
  COEFF *tmppol;
  mpai_t res;
  res->cont=NULL;
  /*init cont to be put here*/
  /*easy cases */
  if(op1->offset<op2->offset)
    {
      if(op1->offset+op1->prec<op2->offset)
	{
	  /*nothing to do*/
	  res->prec=op1->prec;
	  res->offset=op1->offset;
	  tmppol=(COEFF*)MPAI_MALLOC(sizeof(COEFF)*op1->prec);
	  for(i=0;i<op1->prec;i++)
	    gen_init_set(tmppol[i],op1->pol[i]);
	  set_cont(res,op1);/*cont affectation */
	}
      else {
	/*fisrt compute the precision*/ 
	int diff_offset=op2->offset-op1->offset;
	if(op2->offset+op2->prec<op1->offset+op1->prec)
	  res->prec=op2->offset+op2->prec-op1->offset;
	else 
	  res->prec=op1->prec;
	tmppol=(COEFF*)MPAI_MALLOC(res->prec*sizeof(COEFF));
	res->offset=op1->offset;
	for(i=0;i<diff_offset;i++)
	  gen_init_set(tmppol[i],op1->pol[i]);
	for(;i<res->prec;i++)
	  {
	    gen_init(tmppol[i]);
	    gen_add(tmppol[i],op1->pol[i],op2->pol[i-diff_offset]);
	  }
	add_cont(res,op1,op2);
      }
    }
  else
    {
    if(op2->offset+op2->prec<op1->offset)
	{
	  res->prec=op2->prec;
	  res->offset=op2->offset;
	  tmppol=(COEFF*)MPAI_MALLOC(sizeof(COEFF)*op2->prec);
	  for(i=0;i<op1->prec;i++)
	    gen_init_set(tmppol[i],op2->pol[i]);
	  set_cont(res,op2);
	}
      else {
	int diff_offset=op1->offset-op2->offset;
	if(op2->offset+op2->prec>op1->offset+op1->prec)
	  res->prec=op1->offset+op1->prec-op2->offset;
	else 
	  res->prec=op2->prec;
	tmppol=(COEFF*)MPAI_MALLOC(res->prec*sizeof(COEFF));
	res->offset=op2->offset;
	for(i=0;i<diff_offset;i++)
	  gen_init_set(tmppol[i],op2->pol[i]);
	for(;i<res->prec;i++)
	  {
	    gen_init(tmppol[i]);
	    gen_add(tmppol[i],op2->pol[i],op1->pol[i-diff_offset]);
	  }
	    add_cont(res,op2,op1);
      }  
    }
  res->pol=tmppol;
  mpai_set(rop,res);
  mpai_clear(res);
}

void mpai_sub(mpai_ptr rop, mpai_srcptr op1, mpai_srcptr op2)
{
  int i;
  mpai_t res;
  COEFF *tmppol;
  res->cont=NULL;
  if(op1->offset<op2->offset)
    {
      if(op1->offset+op1->prec<op2->offset)
	{
	  res->prec=op1->prec;
	  res->offset=op1->offset;
	  tmppol=(COEFF*)MPAI_MALLOC(sizeof(COEFF)*op1->prec);
	  for(i=0;i<op1->prec;i++)
	    gen_init_set(tmppol[i],op1->pol[i]);
	  set_cont(res,op1);
	}
      else {
	int diff_offset=op2->offset-op1->offset;
	if(op2->offset+op2->prec<op1->offset+op1->prec)
	  res->prec=op2->offset+op2->prec-op1->offset;
	else 
	  res->prec=op1->prec;
	tmppol=(COEFF*)MPAI_MALLOC(res->prec*sizeof(COEFF));
	res->offset=op1->offset;
	for(i=0;i<diff_offset;i++)
	  gen_init_set(tmppol[i],op1->pol[i]);
	for(;i<res->prec;i++)
	  {
	    gen_init(tmppol[i]);
	    gen_sub(tmppol[i],op1->pol[i],op2->pol[i-diff_offset]);
	  }
	sub_cont(res,op1,op2);
      }
    }
  else
    {
    if(op2->offset+op2->prec<op1->offset)
	{
	  res->prec=op2->prec;
	  res->offset=op2->offset;
	  tmppol=(COEFF*)MPAI_MALLOC(sizeof(COEFF)*op2->prec);
	  for(i=0;i<op1->prec;i++)
	    {
	      gen_init_set(tmppol[i],op2->pol[i]);
	    }
	  set_cont(res,op2);
	}
      else {
	int diff_offset=op1->offset-op2->offset;
	if(op2->offset+op2->prec>op1->offset+op1->prec)
	  res->prec=op1->offset+op1->prec-op2->offset;
	else 
	  res->prec=op2->prec;
	tmppol=(COEFF*)MPAI_MALLOC(res->prec*sizeof(COEFF));
	res->offset=op2->offset;
	for(i=0;i<diff_offset;i++)
	  {
	    gen_init(tmppol[i]);
	    gen_sub(tmppol[i],tmppol[i],op2->pol[i]);
	  }
	for(;i<res->prec;i++)
	  {
	    gen_init(tmppol[i]);
	    gen_sub(tmppol[i],op1->pol[i-diff_offset],op2->pol[i]);
	  }
	sub_cont(res,op2,op1);
      }  
    }
  res->pol=tmppol;
  mpai_set(rop,res);
  mpai_clear(res);
}

/*Warning no overlap allowed between input and outputof this function*/
/*Warning 2: Cont function not plugged here to be done */
void mpai_mul(mpai_ptr rop, mpai_srcptr op1, mpai_srcptr op2);
/* karatsuba under suppose prec op1< prec op2 */
void mpai_karatsuba_under_mul(mpai_ptr rop, mpai_srcptr op1, mpai_srcptr op2)
{
  mpai_t res1,res2,res3,accu1,accu2,x1,x0,y1,y0;
  int i,tmp;
  if(op1->prec<=10) {
    COEFF stock;
    gen_init(stock);
    rop->cont=NULL;
    rop->offset=0;
    rop->prec=op1->prec+op2->prec;
    rop->pol=(COEFF*)MPAI_MALLOC(rop->prec*sizeof(COEFF));
    for(i=0;i<rop->prec;i++)
      gen_init(rop->pol[i]);
    for(i=0;i<op1->prec;i++)
      for(tmp=0;tmp<op2->prec;tmp++) {
	gen_mul(stock,op1->pol[i],op2->pol[tmp]);
	gen_add(rop->pol[i+tmp],rop->pol[i+tmp],stock);
      }
    gen_clear(stock);
  }
  else {
    rop->offset=0;
    rop->prec=(op1->prec+op2->prec);
    rop->pol=(COEFF*)MPAI_MALLOC(rop->prec*sizeof(COEFF));
    rop->cont=NULL;
    for(i=0;i<rop->prec;i++)
      gen_init(rop->pol[i]);
    *x0=*x1=*op1;
    x0->offset=x1->offset=0;
    *y0=*y1=*op2;
    y0->offset=y1->offset=0;
    x0->prec=op1->prec/2;
    x1->prec=op1->prec-x0->prec;
    x1->pol+=x0->prec;
    y0->prec=x0->prec;
    y1->prec=op2->prec-x0->prec;
    y1->pol+=x0->prec;
    mpai_karatsuba_under_mul(res1,x0,y0);
    mpai_karatsuba_under_mul(res2,x1,y1);
    accu1->cont=NULL;
    accu1->prec=x1->prec;
    accu1->pol=rop->pol;
    accu1->offset=0;
    for(i=0;i<x0->prec;i++) {
      gen_sub(accu1->pol[i],x1->pol[i],x0->pol[i]);
    }
    for(;i<x1->prec;i++)
      gen_set(accu1->pol[i],x1->pol[i]);
    accu2->prec=y1->prec;
    accu2->pol=rop->pol+2*x0->prec;
    accu2->offset=0;
    accu2->cont=NULL;
    for(i=0;i<y0->prec;i++) {
      gen_sub(accu2->pol[i],y1->pol[i],y0->pol[i]);
    }
    for(;i<y1->prec;i++)
      gen_set(accu2->pol[i],y1->pol[i]);
    if(accu1->prec<accu2->prec)
      mpai_karatsuba_under_mul(res3,accu1,accu2);
    else
      mpai_karatsuba_under_mul(res3,accu2,accu1);
    for(i=0;i<res1->prec;i++)
      gen_set(rop->pol[i],res1->pol[i]);
    for(;i<accu1->prec;i++)
      gen_set_d(rop->pol[i],0);
    for(i=0;i<res2->prec;i++)
      gen_set(rop->pol[i+2*x0->prec],res2->pol[i]);
    for(i=0;i<res3->prec;i++)
      gen_sub(rop->pol[i+x0->prec],rop->pol[i+x0->prec],res3->pol[i]);
    for(i=0;i<res1->prec;i++)
      gen_add(rop->pol[i+x0->prec],rop->pol[i+x0->prec],res1->pol[i]);
    for(i=0;i<res2->prec;i++)
      gen_add(rop->pol[i+x0->prec],rop->pol[i+x0->prec],res2->pol[i]);
    mpai_clear(res1);
    mpai_clear(res2);
    mpai_clear(res3);
  }
}

void mpai_karatsuba_mul(mpai_ptr rop, mpai_srcptr op1, mpai_srcptr op2)
{
  mpai_t res1;
  int i;
  if(op1->prec<op2->prec) {
      mpai_karatsuba_under_mul(res1,op1,op2); 
      for(i=op1->prec;i<res1->prec;i++)
	gen_clear(res1->pol[i]);
      for(i=0;i<rop->prec;i++)
	gen_clear(rop->pol[i]);
      MPAI_FREE(rop->pol);
      rop->pol=(COEFF*)MPAI_REALLOC(res1->pol,res1->prec*sizeof(coeff)
				  ,op1->prec*sizeof(COEFF));
      rop->prec=op1->prec;
  }
  else {
    mpai_karatsuba_under_mul(res1,op2,op1);
      for(i=op2->prec;i<res1->prec;i++)
	gen_clear(res1->pol[i]);
      for(i=0;i<rop->prec;i++)
	gen_clear(rop->pol[i]);
      MPAI_FREE(rop->pol);
      rop->pol=(COEFF*)MPAI_REALLOC(res1->pol,res1->prec*sizeof(coeff)
				  ,op2->prec*sizeof(COEFF));
    rop->prec=op2->prec;
  }
  rop->offset=op1->offset+op2->offset;
}

void mpai_short_prod(mpai_ptr rop, mpai_srcptr op1, mpai_srcptr op2)
{
  mpai_t res,x0,x1,res2,y0,y1,res3;
  int i,k,tmp;
    tmp=min(op1->prec,op2->prec);
  if(tmp==1) {
      int tmpprec;
      mpai_clear(rop);
      tmpprec=mpai_prec;
      mpai_set_prec(1);
      mpai_init(rop);
      mpai_set_prec(tmpprec);
      gen_mul(rop->pol[0],op1->pol[0],op2->pol[0]);
      rop->offset=op1->offset+op2->offset;
    }
  else {
    int stock;
     for(stock=0;tmp>>stock >0;stock++);
    stock--;
    k= 1 << (int)stock;
    mpai_init(res2);
    mpai_init(res3);
    *x0=*x1=*op1;
    *y0=*y1=*op2;
    x0->prec=k;
    y0->prec=k;
    x0->offset=x1->offset=y1->offset=y0->offset=0;
    mpai_karatsuba_under_mul(res,x0,y0);
    for(i=tmp;i<res->prec;i++)
      gen_clear(res->pol[i]);
    res->pol=(COEFF*)MPAI_REALLOC(res->pol,res->prec*sizeof(COEFF)
				  ,tmp*sizeof(COEFF));
    for(i=res->prec;i<tmp;i++)
      gen_init(res->pol[i]);
    res->prec=tmp;
    res->cont=NULL;
    res->offset=op1->offset+op2->offset;
    y1->pol+=k;
    x1->pol+=k;
    x1->prec=tmp-k;
    y1->prec=tmp-k;
    y0->prec=tmp-k;
    x0->prec=tmp-k;
    if(tmp!=k)
      {
	mpai_short_prod(res2,x1,y0);
	mpai_short_prod(res3,x0,y1);
      }
    for(i=0;i<tmp-k;i++)
      gen_add(res->pol[i+k],res->pol[i+k],res2->pol[i]);
    for(i=0;i<tmp-k;i++)
      gen_add(res->pol[i+k],res->pol[i+k],res3->pol[i]);
    mpai_clear(rop);
    *rop=*res;
    mpai_clear(res2);
    mpai_clear(res3);
  }
}

void mpai_mul(mpai_ptr rop,mpai_srcptr op1,mpai_srcptr op2)
{
  int i,j;
  mpai_t res;
  COEFF *pol,tmp;
  gen_init(tmp);
  res->cont=NULL;
  res->offset=op1->offset+op2->offset;
  res->prec=min(op1->prec,op2->prec);
  pol=(COEFF*)MPAI_MALLOC(res->prec*sizeof(COEFF));
  for(i=0;i<res->prec;i++)
    gen_init(pol[i]);
  for(i=0;i<res->prec;i++)
    for(j=0;j+i<res->prec;j++)
      {
	gen_mul(tmp,op1->pol[i],op2->pol[j]);
	gen_add(pol[i+j],pol[i+j],tmp);
      }
  res->pol=pol;
  mul_cont(res,op1,op2);
  mpai_set(rop,res);
  mpai_clear(res);
  gen_clear(tmp);
}
#ifdef ESSAI
void mpai_karait_mul(mpai_ptr rop,mpai_srcptr op1, mpai_srcptr op2)
{
  mpai_t res1,res2,res3,accu1,accu2,x1,x0,y1,y0;
  int i,tmp,tmpprec;
  //  mpai_nbmul++;
  if(op1->prec<=10) {
    //    mpai_out_str(stdout,op2);
    //printf("avant mul \n");
    COEFF stock;
    gen_init(stock);
    rop->cont=NULL;
    rop->offset=0;
    rop->prec=op1->prec+op2->prec;
    rop->pol=(COEFF*)MPAI_MALLOC(rop->prec*sizeof(COEFF));
    for(i=0;i<rop->prec;i++)
      gen_init(rop->pol[i]);
    for(i=0;i<op1->prec;i++)
      for(tmp=0;tmp<op2->prec;tmp++) {
	gen_mul(stock,op1->pol[i],op2->pol[tmp]);
	gen_add(rop->pol[i+tmp],rop->pol[i+tmp],stock);
      }
    gen_clear(stock);
  }
  else {
    rop->offset=0;
    tmpprec=(op1->prec+op2->prec);
    rop->prec=tmpprec;
    rop->pol=(COEFF*)MPAI_MALLOC(rop->prec*sizeof(COEFF));
    rop->cont=NULL;
    for(i=0;i<rop->prec;i++)
      gen_init(rop->pol[i]);
    
    //attention ne pas utiliser les mpai_operations
    //elle marchent sur des series tronquees !
    //et ne calculent pas la somme totale!
    //Calcul du nombre de passes
    for(tmp=0;op1->prec >>tmp >0;tmp++);
    //init de rop
    for(i=0;i<op1->prec;i++)
      gen_mul(rop->pol[i],op1->pol[i],op2->pol[i]);
    //maintenant on fait les passes
    for(i=1;i<tmp;i++) {
      *x0=*x1=*op1;
      x1->pol+=1<<(i-1);
      *y0=*y1=*op2;
      y1->pol+=1<<(i-1);
      x0->prec=y0->prec=1<<(i-1);
      for(j=0;j<tmprec;j+=1<<i){
	  
	}
      }
}
#endif 

void mpai_nonewton_div(mpai_ptr rop, mpai_srcptr op1, mpai_srcptr op2)
{
 int i;
 COEFF stock;
 mpai_t temp,accu,inv,res;
 res->offset=op1->offset-op2->offset;
 res->prec=min(op1->prec,op2->prec);
 mpai_init(temp);
 mpai_init(accu);
 mpai_init(inv);
 mpai_set(temp,op2);
 temp->prec--;
 gen_init(stock);
 gen_set(stock,*(temp->pol));
 for(i=1;i<=temp->prec;i++)
   {
     gen_set(temp->pol[i-1],temp->pol[i]);
     gen_div(temp->pol[i-1],temp->pol[i-1],stock);
     gen_neg(temp->pol[i-1],temp->pol[i-1]);
   }
 gen_clear(temp->pol[temp->prec]);
 temp->offset++;
 mpai_set_si(inv,1);
 mpai_set_si(accu,1);
 for(i=1;i<res->prec;i++)
   {
     mpai_mul(accu,accu,temp);
     mpai_add(inv,inv,accu);
   }
 mpai_div_d(inv,op2->pol[0]);
 mpai_mul(rop,op1,inv);
 gen_clear(stock);
 mpai_clear(inv);
 mpai_clear(accu);
 mpai_clear(temp);
 div_cont(rop,op1,op2);
}

void mpai_div(mpai_ptr rop, mpai_srcptr op1, mpai_srcptr op2)
{
  mpai_t inverse, stock, un;
  int tmp=op2->prec,i;
  for(tmp=0;op2->prec>>tmp >0;tmp++);
  mpai_init(inverse);
  mpai_init(un);
  mpai_init(stock);
  gen_set_d(un->pol[0],1);
  gen_set_d(inverse->pol[0],1);
  gen_div(inverse->pol[0],inverse->pol[0],op2->pol[0]);
  for(i=0;i<tmp;i++){
    mpai_short_prod(stock, inverse, op2);
    mpai_sub(stock,un,stock);
    mpai_short_prod(stock,inverse,stock);
    mpai_add(inverse,inverse,stock);
  }
  mpai_short_prod(rop,op1,inverse);
  mpai_clear(stock);
  mpai_clear(inverse); 
  mpai_clear(un);
  div_cont(rop,op1,op2);
}

void mpai_out_str(FILE* stream,mpai_srcptr op)
{
  int i=0;
  for(;i<op->prec-1;i++)
    {
      gen_out_stream(stream,op->pol[i]);
      fprintf(stream,"*x^%d+",i+op->offset);
    }
  gen_out_stream(stream,op->pol[i]);
  fprintf(stream,"*x^%d",i+op->offset);
}

void mpai_mul_eps(mpai_ptr res,const int exp)
{
  res->offset+=exp;
}

int mpai_cmp(mpai_srcptr op1, mpai_srcptr op2)
{
  int minprec=min(op1->prec,op2->prec);
  int i,res=0;
  for(i=0;i<minprec;i++)
    if ((res=gen_cmp(op1->pol[i],op2->pol[i])))
      break;
  if(op1->prec>op2->prec)
    {res=gen_sgn(op1->pol[i]);}
  if(op1->prec<op2->prec)
    {res=gen_sgn(op2->pol[i]);}
  return res;
}

int mpai_cmp_ui(mpai_srcptr op1, const RS_I32 op2)
{
  int res=1;
  mpai_t tmp;
  mpai_init(tmp);
  mpai_set_si(tmp,op2);
  res=mpai_cmp(op1,tmp);
  mpai_clear(tmp);
  return res;
}

int mpai_isNull(mpai_srcptr op)
{
  int i,res=1;
  COEFF tmp;
  gen_init(tmp);
  for(i=0;i<op->prec;i++)
    if(gen_cmp(op->pol[i],tmp)!=0)
      {
	res=0;
	break;
      }
  gen_clear(tmp);
  return res;
}

int mpai_isPos(mpai_srcptr op)
{
  int i,res=0;
  COEFF tmp;
  gen_init(tmp);
  gen_set_d(tmp,0);
  for(i=0;i<op->prec && (gen_cmp(op->pol[i],tmp)==0);i++);
  if(i!=op->prec)
    res=(gen_cmp(op->pol[i],tmp)>0);
  return res;
}

int mpai_isNeg(mpai_srcptr op)
{
  return !mpai_isPos(op);
}

int mpai_isUnit(mpai_srcptr op)
{
  int i,res=0;
  COEFF tmp;
  gen_init(tmp);
  gen_set_d(tmp,0);
  for(i=0;i<op->prec && !gen_cmp(op->pol[i],tmp);i++);
  gen_set_d(tmp,1);
  if(i<op->prec)
    res=(gen_cmp(op->pol[i],tmp) && !(i+op->offset));
  gen_clear(tmp);
  return res;
}

int mpai_isEq(mpai_srcptr op1, mpai_srcptr op2)
{
  return mpai_cmp(op1,op2)==0;
}

void mpai_exactdiv(mpai_ptr rop, mpai_srcptr op1, mpai_srcptr op2)
{
  mpai_div(rop,op1,op2);
}

void mpai_addmul(mpai_ptr rop, mpai_srcptr op1,mpai_srcptr op2,mpai_srcptr op3)
{
  mpai_t tmp;
  mpai_init(tmp);
  mpai_mul(tmp,op1,op2);
  mpai_add(rop,tmp,op3);
  mpai_clear(tmp);
}

void mpai_submul(mpai_ptr rop, mpai_srcptr op1,mpai_srcptr op2,mpai_srcptr op3)
{
  mpai_t tmp;
  mpai_init(tmp);
  mpai_mul(tmp,op1,op2);
  mpai_sub(rop,tmp,op3);
  mpai_clear(tmp);
}

void mpai_comb(mpai_ptr rop, mpai_srcptr op1,mpai_srcptr op2,
	       mpai_srcptr op3, mpai_srcptr op4)
{
  mpai_t tmp,tmp2;
  mpai_init(tmp);
  mpai_init(tmp2);
  mpai_mul(tmp,op1,op2);
  mpai_mul(tmp2,op3,op4);
  mpai_add(rop,tmp,tmp2);
  mpai_clear(tmp);
  mpai_clear(tmp2);
}

void mpai_neg(mpai_ptr rop, mpai_srcptr op)
{
  int i;
  mpai_set(rop,op);
  for(i=0;i<rop->prec;i++)
    gen_neg(rop->pol[i],rop->pol[i]);
}

void mpai_inv(mpai_ptr rop, mpai_srcptr op)
{
  mpai_t tmp;
  mpai_init(tmp);
  mpai_set_si(tmp,1);
  mpai_div(rop,tmp,op);
}
  
void mpai_pow(mpai_ptr rop, mpai_srcptr op, const long c)
{}

void mpai_gcd(mpai_ptr rop, mpai_srcptr op1, mpai_srcptr op2)
{}

void mpai_rem(mpai_ptr rop, mpai_srcptr op1, mpai_srcptr op2)
{}

void mpai_lcm(mpai_ptr rop, mpai_srcptr op1, mpai_srcptr op2)
{}

long mpai_mod(mpai_srcptr op,const long b)
{
  return 0;
}

void mpai_dadd(mpai_ptr op1, mpai_srcptr op2)
{
  mpai_add(op1,op1,op2);
}

void mpai_dsub(mpai_ptr op1, mpai_srcptr op2)
{
  mpai_sub(op1,op1,op2);
}

void mpai_dmul(mpai_ptr op1, mpai_srcptr op2)
{
  mpai_mul(op1,op1,op2);
}

void mpai_ddiv(mpai_ptr op1, mpai_srcptr op2)
{
  mpai_div(op1,op1,op2);
}

void mpai_dexactdiv(mpai_ptr op1, mpai_srcptr op2)
{
  mpai_div(op1,op1,op2);
}

void mpai_daddmul(mpai_ptr op1, mpai_srcptr op2,mpai_srcptr op3)
{
  mpai_addmul(op1,op1,op2,op3);
}

void mpai_dsubmul(mpai_ptr op1, mpai_srcptr op2,mpai_srcptr op3)
{
  mpai_submul(op1,op1,op2,op3);
}

void mpai_dcomb(mpai_ptr op1, mpai_srcptr op2,
	       mpai_srcptr op3, mpai_srcptr op4)
{
  mpai_comb(op1,op1,op2,op3,op4);
}

void mpai_dneg(mpai_ptr op)
{
  mpai_neg(op,op);
}

void mpai_dinv(mpai_ptr op)
{
  mpai_inv(op,op);
}

void mpai_dpow(mpai_ptr rop,const long u)
{}

void mpai_dgcd(mpai_ptr rop,mpai_srcptr op)
{}

void mpai_drem(mpai_ptr rop, mpai_srcptr op)
{}

void mpai_dlcm(mpai_ptr rop, mpai_srcptr op)
{}

